/* gist.c -- GIST main program, version 0.3

   GIST will carry out the genotype-IBD sharing test described in
       Li et al. (2004) Assessing Whether an Allele Can Account
       in Part for a Linkage Signal: The Genotype-IBD Sharing Test (GIST).
       Am. J. Hum. Genet. 74:418-431
   More information can be found in the readme file that comes with this
   program.

   Syntax:
       command file1 [file2]
   file1 -- name of the input file.
   file2 -- name of the output file.  If omitted, the default is stdout.

   Chun Li, 03/24/2004, chun.li@vanderbilt.edu
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "CORR.h"

#define SQUARE(a) (a)*(a)
#define NEMP 100000   /* number of data points in empirical distribution */
#define MAXBUF 1048576  /* maximum size of input file */
#define MAXLINE 1024  /* maximum number of characters per line in input file */
#define MAXFAM 5000   /* maximum number of families in input file */

char conbuf[MAXBUF], line[MAXLINE];
long ichcon;

FILE *in, *out;

int nfam=0;
double REVNEMP, FACTOR;
double allfreq=0.5;

double wt_aver1, wt_aver2, wt_aver3, revsd1, revsd2, revsd3;
double wt_sum1, wt_sum2, wt_sum3, wt_sumsq1, wt_sumsq2, wt_sumsq3; 
double wtdt_sum1, wtdt_sum2, wtdt_sum3;
double npl_sum, ibd_sumsq, revibdsd;

double *weight1, *weight2, *weight3, *npl;
double truest1, truest2, truest3, truest4, pv1, pv2, pv3, pv4;
double wnpl1, wnpl2, wnpl3;

double *max, *MAX;
int *empidx;


int iseed1=16113, iseed2=5838, iseed3=10758;
int it1[32767], it2[32767], it3[32767];

void init();
double randm(void); /* stdlib.h has already defined a function random */

void indexx(double *array, int *irank, int n);
double gausscdf(double x);
double gaussgen();

void replica();
void precalc();
double wnpl(double *data2, double *weight);
int getline(char *pinbuf, long *ichin);

void empirical(double p);

void incon();
void header();
void alter();
void results();
void allocate();
void gisthelp();



int main(int argc, char *argv[])
{
  int i;
  out=stdout;

  /* Read in paremeters from the control file. */
  if(argc<2) {
    gisthelp();
    return 0;
  }
  if((in=fopen(argv[1], "r"))==NULL) {
    printf("Cannot open file %s to read\n", argv[1]);
    return 1;
  }

  if(argc>2)
    if((out=fopen(argv[2], "w"))==NULL) {
      printf("Cannot open file %s to write\n", argv[2]);
      return 1;
    }


  /* Initialize random number generator, allocate memory, and output
     relevant information. */
  init(it1, it2, it3);
  allocate();

  incon();
  fclose(in);

  FACTOR=sqrt(nfam-3);
  REVNEMP=1.0/NEMP;

  empirical(allfreq);
  precalc();
  replica();

  fprintf(out, "\nGIST Analysis, version 0.3\n\n");
  fprintf(out, "Number of families analyzed: %d\n\n", nfam);

  fprintf(out, "Test               P-value      Weighted NPL score\n\n");
  fprintf(out, "Dominant model     %8.6f     %6.4f\n", pv1, wnpl1);
  fprintf(out, "Recessive model    %8.6f     %6.4f\n", pv2, wnpl2);
  fprintf(out, "Additive model     %8.6f     %6.4f\n", pv3, wnpl3);
  fprintf(out, "Combined (Max)     %8.6f\n\n", pv4);

  fprintf(out, "Note:  All tests are one-sided tests.\n");
  fprintf(out, "Note:  Weighted NPL scores should not be interpreted the same as\n a regular NPL score.\n\n");

  fclose(out);
  return 0;
}


/*  Generate an index array (irank) for an numeric array (array).
    The index runs from 0 to n-1.
*/
void indexx(double *array, int *irank, int n)
{
  int i, j, l, ir, irankt, *prank;
  double q;

  prank=irank;
  for (j=0; j<n; j++) {
    *prank++=j;
  }

  l=n/2;
  ir=n-1;

  for(;;) {
    if (l > 0) {
      irankt=irank[--l];
      q=array[irankt];
    }
    else {
      irankt=irank[ir];
      q=array[irankt];
      irank[ir]=irank[0];
      if (!(--ir)) {
        irank[0]=irankt;
        return;
      }
    }

    i=l;
    j=l+l+1;

    while (j <= ir) {
      if (j < ir && array[irank[j]] < array[irank[j+1]])
        j++;
      if (q < array[irank[j]]) {
        irank[i]=irank[j];
        i=j;
        j+=j+1;
      }
      else {
        j=ir+1;
      }
    }

    irank[i]=irankt;
  }
}


/* Computes the standard normal distribution function at the point x.
   For x between -4. and + 4, the power series expansion of Polya is used.
   For x outside this range, the asymptotic expansion for Mills' ratio is
   employed.  See: M.G. Kendall and A. Stuart(1963). "The Advanced Theory
   of Statistics, vol. 1, 2nd ed." Hafner Publishing. New York.
*/
double gausscdf(double x)
{
  double xsq, gaucdf, t, ratio, r;
  /* factor is the reciprocal of the square root of 2*pi */
  const double factor=0.3989422804014327;

  xsq=x*x;
  gaucdf=0;
  r=1;

  if(xsq<=16) {
    t=factor*exp(-xsq/2);
    do {
      gaucdf+=t;
      r+=2;
      ratio=xsq/r;
      t*=ratio;
    } while(t > 1e-10 || ratio > 0.9);
    return gaucdf*x+0.5;
  }

  else {
    t=factor*exp(-xsq/2)/abs(x);
    do {
      gaucdf+=t;
      r+=2;
      ratio=r/xsq;
      t=-t*ratio;
    } while (ratio < 1 && abs(t) > 1e-10);
    return x>0 ? 1-gaucdf : gaucdf;
  }
}


/*  These two functions require the main program declare and initialize
        int iseed1, iseed2, iseed3;
    and declare arrays
        int it1[32767], it2[32767], it3[32767];
    preferably as global variables.
*/

/* Initiate the Wichmann-Hill random number generator.
   This has already been tested to yield the same values as the
   Wichmann-Hill algorithm (JRSS Series C 1982 31:188-190, with
   a minor correction in JRSS Series C 1984 33:123).  The period
   of this algorithm is 6.95e12.
*/
void init()
{
  register int i;

  for(i=0; i<30323; i++) {
    it1[i]=(171*(i%177)-2 *(i/177)+30269)%30269;
    it2[i]=(172*(i%176)-35*(i/176)+30307)%30307;
    it3[i]=(170*(i%178)-63*(i/178)+30323)%30323;
  }
}

/* The Wichmann-Hill random number generator.  */
double randm()
{
  double x;
  
  iseed1=it1[iseed1];
  iseed2=it2[iseed2];
  iseed3=it3[iseed3];

/* The constants are 1/30269.0, 1/30307.0, 1/30323.0 */
  x = iseed1 * 3.3037100664045725e-05 +
    iseed2 * 3.2995677566238825e-05 +
    iseed3 * 3.297826732183491e-05;

  return x-(int)x;
}


/* This standard normal random generator uses the modified Box-Miller method.
   This modification throws out 21.46% of the data but totally avoids the
   functional calls to sin and cos.  The time spent is about 2/3 of that for
   the direct implementation of the Box-Miller method.
   Press et al. (1992)
 */
double gaussgen()
{
  double v1, v2, rsq, fac;
  static double reserve;
  static int i=0;

  if(i) {
    i=0;
    return reserve;
  }
  else {
    do {
      v1=randm()*2-1;
      v2=randm()*2-1;
      rsq=v1*v1+v2*v2;  /* rsq is uniform */
    } while (rsq>=1 || rsq==0);

    fac=sqrt(-2*log(rsq)/rsq);
    reserve=v2*fac;
    i=1;
    return v1*fac;
  }
}


void replica()
{
  int j, *p1;
  double *p2, tmp;

  truest1= (wtdt_sum1 - wt_aver1*npl_sum) * revsd1 * revibdsd;
  truest2= (wtdt_sum2 - wt_aver2*npl_sum) * revsd2 * revibdsd;
  truest3= (wtdt_sum3 - wt_aver3*npl_sum) * revsd3 * revibdsd;

  truest1=0.5*log(2/(1-truest1)-1)*FACTOR;
  truest2=0.5*log(2/(1-truest2)-1)*FACTOR;
  truest3=0.5*log(2/(1-truest3)-1)*FACTOR;

  pv1=1-gausscdf(truest1);
  pv2=1-gausscdf(truest2);
  pv3=1-gausscdf(truest3);

  /* max */
  tmp = truest1 > truest2 ? truest1 : truest2;
  truest4= tmp > truest3 ? tmp : truest3;

  p2=MAX;
  for(j=0; j<NEMP; j++)
    if(truest4 > *(p2++)) break;
  pv4= REVNEMP*j;

  wnpl1=wnpl(npl, weight1);
  wnpl2=wnpl(npl, weight2);
  wnpl3=wnpl(npl, weight3);
}


void precalc()
{
  int i;
  double *pnpl, *pw1, *pw2, *pw3, tnpl, tw1, tw2, tw3;
  double sumnpl, sumw1, sumw2, sumw3, sqnpl, sqw1, sqw2, sqw3, prod1, prod2, prod3;

  pnpl=npl;
  pw1=weight1;
  pw2=weight2;
  pw3=weight3;

  sumnpl=sumw1=sumw2=sumw3=0;
  sqnpl=sqw1=sqw2=sqw3=0;
  prod1=prod2=prod3=0;

  for(i=0; i<nfam; i++) {
    tnpl=*(pnpl++);
    tw1 =*(pw1++);
    tw2 =*(pw2++);
    tw3 =*(pw3++);

    sumnpl+= tnpl;
    sumw1+= tw1;
    sumw2+= tw2;
    sumw3+= tw3;

    sqnpl+= tnpl*tnpl;
    sqw1+= tw1*tw1;
    sqw2+= tw2*tw2;
    sqw3+= tw3*tw3;

    prod1+= tnpl*tw1;
    prod2+= tnpl*tw2;
    prod3+= tnpl*tw3;
  }

  wt_aver1=sumw1/nfam;
  wt_aver2=sumw2/nfam;
  wt_aver3=sumw3/nfam;
  npl_sum=sumnpl;

  wtdt_sum1=prod1;
  wtdt_sum2=prod2;
  wtdt_sum3=prod3;

  revsd1=1.0/sqrt(sqw1-sumw1*wt_aver1);
  revsd2=1.0/sqrt(sqw2-sumw2*wt_aver2);
  revsd3=1.0/sqrt(sqw3-sumw3*wt_aver3);
  revibdsd=1.0/sqrt(sqnpl-sumnpl*sumnpl/nfam);
}


double wnpl(double *data2, double *weight)
{
  int i;
  double *pd, *pw, a, b;

  pd=data2;
  pw=weight;
  a=b=0;
  for(i=0; i<nfam; i++) {
    a+=(*pw)*(*pd);
    b+=(*pw)*(*pw);
    pd++; pw++;
  }

  return a/sqrt(b);
}


void incon()
{
  int chcon, n;

  /* This is used to catch lines in the input with >4 numbers. */
  char junk[1024];

  chcon=fread(conbuf, 1, MAXBUF, in);
  if(chcon==0)
    {
      printf("Error:  Input file is empty.\n\n");
      exit(1);
    }
  ichcon=0L;

  getline(conbuf, &ichcon);
  sscanf(line, "%lf", &allfreq);

  do {
    getline(conbuf, &ichcon);
    n=sscanf(line, "%lf %lf %lf %lf %s",
	     &weight1[nfam], &weight2[nfam], &weight3[nfam], &npl[nfam], junk);
    if(n==0 || n==EOF) break;
    if(n!=4) {
      if(n>4) printf("Error:  Line %d of the input file has >4 numbers.\n\n", nfam+2);
      else printf("Error:  Line %d of the input file has <4 numbers.\n\n", nfam+2);
      exit(1);
    }

    if(++nfam > MAXFAM)
      {
	printf("Error:  Number of families exceeds the maximum allowed (%d).\nTo allow more families, please modify the source code and recompile.\n\n", MAXFAM);
	exit(1);
      }
  } while (ichcon<chcon);
}


int getline(char *pinbuf, long *ichin)
{
  int i=0;
  char *pline, *pichin;

  pline=line;
  pichin=pinbuf+*ichin;

  do {
    *pline++=*pichin;
    if(++i >= MAXLINE)
      {
	printf("Error:  Number of characters per line exceeds the maximum allowed (%d).\nTo allow more families, please modify the source code and recompile.\n\n", MAXLINE);
	exit(1);
      }
    (*ichin)++;
    
  } while (*pichin++ != '\n');

  /* This ensures the line is terminated correctly. */
  *pline='\000';

  return;
}


void empirical(double p)
{
  int i, ip;
  double rho12, rho13, rho23, a, b, x, y1, y2, y3, tmp;

  ip=(int)(p*100+0.5);
  if(ip<3 || ip>97) {
    printf("Allele frequency %f is too extreme\n", p);
    exit(1);
  }

  rho12=corr[ip][1];
  rho13=corr[ip][2];
  rho23=corr[ip][3];

  a=sqrt(1-rho12*rho12);
  b=(rho23-rho12*rho13)/a;

  for(i=0; i<NEMP; i++) {
    x=gaussgen();
    y1=gaussgen();
    y2=rho12*y1 + a*x;
    y3=rho13*y1 + b*x + sqrt(1-rho13*rho13-b*b)*gaussgen();

    tmp = y1 > y2 ? y1 : y2;
    max[i] = tmp > y3 ? tmp : y3;
  }

  indexx(max, empidx, NEMP);
  for(i=0; i<NEMP; i++)
    MAX[NEMP-1-i]=max[empidx[i]];
}


void allocate()
{
  npl=(double *)malloc((MAXFAM+1)*sizeof(double));
  weight1=(double *)malloc((MAXFAM+1)*sizeof(double));
  weight2=(double *)malloc((MAXFAM+1)*sizeof(double));
  weight3=(double *)malloc((MAXFAM+1)*sizeof(double));

  max=(double *)malloc(NEMP*sizeof(double));
  MAX=(double *)malloc(NEMP*sizeof(double));
  empidx=(int *)malloc(NEMP*sizeof(int));
}


void gisthelp()
{
  printf("GIST, version 0.3\n\n");
  printf("Syntax:\n");
  printf("    gist.exe inputfile [outputfile]\n\n");

  printf("If the output file name is omitted, the standard output will be used.\n");
  printf("Please see the readme file for the format of input file.\n\n");

  printf("GIST is developed and maintained by Dr. Chun Li of Vanderbilt University.\n");
  printf("Please send your comments and suggestions to chun.li@vanderbilt.edu.\n");

}
